@@ -9,9 +9,11 @@
 
 expm.s.Pade.s <- function(x, order, n=nrow(x)) {
     ## no checking here; this is not to be called by the user
+    ## try have this work with "mpfrMatrix" <==> solve(<mpfrArray>)
 
+    ## s := the number of [s]quarings
     e <- ceiling(log2(max(rowSums(abs(x)))))
-    s <- max(0, e+1)
+    s <- max(e+1, 0)
 
     ## preconditions x :
     x <- x / (2^s)
@@ -41,9 +43,11 @@ expm.s.Pade.s <- function(x, order, n=nrow(x)) {
 
 
 expm.methSparse <- c("Higham08", "R_Eigen", "R_Pade")
-## "FIXME" -- keep this list up-to-date - test by setting  R_EXPM_NO_DENSE_COERCION
+## keep this list up-to-date - test by setting  R_EXPM_NO_DENSE_COERCION
+## but NOTE: It may make sense to *keep* the message() about coercion to dense (memory blow up!)
 
 expm <- function(x, method = c("Higham08.b", "Higham08",
+                    "AlMohy-Hi09",
 		    "Ward77", "PadeRBS", "Pade", "Taylor", "PadeO", "TaylorO",
 		    "R_Eigen", "R_Pade", "R_Ward77", "hybrid_Eigen_Ward"),
 		 order = 8,
@@ -51,30 +55,33 @@ expm <- function(x, method = c("Higham08.b", "Higham08",
 		 do.sparseMsg = TRUE,
 		 preconditioning = c("2bal", "1bal", "buggy"))
 {
-    ## some methods work for "matrix" or "Matrix" matrices:
-    stopifnot(is.numeric(x) || is(x, "dMatrix"),
-	      length(d <- dim(x)) == 2)
+    ## some methods work for "matrix", "Matrix", or "mpfrMatrix", iff solve(.,.) worked:
+    stopifnot(is.numeric(x) || (isM <- inherits(x, "dMatrix")) || inherits(x, "mpfrMatrix"))
+    if(length(d <- dim(x)) != 2) stop("argument is not a matrix")
     if (d[1] != d[2]) stop("matrix not square")
     method <- match.arg(method)
     checkSparse <- !nzchar(Sys.getenv("R_EXPM_NO_DENSE_COERCION"))
-    if(!is.numeric(x) && checkSparse) { # i.e., a "dMatrix"
+    isM <- !is.numeric(x) && isM
+    if(isM && checkSparse) { # i.e., a "dMatrix"
 	if(!(method %in% expm.methSparse) && is(x, "sparseMatrix")) {
 	    if(do.sparseMsg)
-		message("coercing to dense matrix, as required by method ",
-			dQuote(method))
+		message(gettextf("coercing to dense matrix, as required by method %s",
+                                 dQuote(method)), domain=NA)
 	    x <- as(x, "denseMatrix")
 	}
     }
     switch(method,
+           "AlMohy-Hi09" = expm.AlMoHi09(x, p = order)
+          ,
 	   "Higham08.b" = expm.Higham08(x, balancing = TRUE)
-	   ,
+	  ,
 	   "Higham08"	= expm.Higham08(x, balancing = FALSE)
-	   ,
+          ,
 	   "Ward77" = {
 	       ## AUTHORS: Christophe Dutang, Vincent Goulet at act ulaval ca
 	       ##	 built on "Matrix" package, built on 'octave' code
 	       ##	 Martin Maechler, for the preconditioning etc
-               stopifnot(is.matrix(x))
+               if(!is.numeric(x)) x <- as(x, "matrix")
 	       switch(match.arg(preconditioning),
 		      "2bal" = .Call(do_expm, x, "Ward77"),
 		      "1bal" = .Call(do_expm, x, "Ward77_1"),
@@ -98,8 +105,8 @@ expm <- function(x, method = c("Higham08.b", "Higham08",
 	       ## AUTHOR: Christophe Dutang
 	       ## matrix exponential using eigenvalues / spectral decomposition and
 	       ## Ward(1977) algorithm if x is numerically non diagonalisable
-               stopifnot(is.matrix(x))
-	       .Call("do_expm_eigen", x, tol)
+               if(!is.numeric(x)) x <- as(x, "matrix")
+	       .Call(do_expm_eigen, x, tol)
 	   },
 	   "R_Pade"= { ## use scaling + Pade + squaring with R code:
 
@@ -163,9 +170,9 @@ expm <- function(x, method = c("Higham08.b", "Higham08",
 	       else x
 	   },
 	   "PadeRBS" =  { ## the "expofit" method by  Roger B. Sidje (U.Queensland, AU)
-	       stopifnot(is.matrix(x),
-			 (order <- as.integer(order)) >= 1)
-	       storage.mode(x) <- "double"
+               if(!is.numeric(x)) x <- as(x, "matrix")
+	       stopifnot((order <- as.integer(order)) >= 1)
+	       if(!is.double(x)) storage.mode(x) <- "double"
 	       Fobj <- .Fortran(matexpRBS,
 				order,		   # IDEG  1
 				as.integer(d[1]),  # M	   2
@@ -179,15 +186,22 @@ expm <- function(x, method = c("Higham08.b", "Higham08",
            }
 	   , { ## the "mexp" methods by
 	       ## AUTHORS: Marina Shapira and David Firth --------------
-               stopifnot(is.matrix(x))
-	       storage.mode(x) <- "double"
+               if(!is.numeric(x)) x <- as(x, "matrix")
+	       if(!is.double(x)) storage.mode(x) <- "double"
 	       order <- as.integer(order)
 	       ## MM:	a "silly"  way to code the method / order
 	       ntaylor <- npade <- 0L
 	       if (substr(method,1,4) == "Pade")
 		   npade <- order else ntaylor <- order
-	       res <- .Fortran(if(identical(grep("O$", method), 1L))
-			       matrexpO else matrexp,
+	       res <- if(identical(grep("O$", method), 1L))
+			  .Fortran(matrexpO,
+			       X = x,
+			       size = d[1],
+			       ntaylor,
+			       npade,
+			       accuracy = double(1))[c("X", "accuracy")]
+		      else
+			  .Fortran(matrexp,
 			       X = x,
 			       size = d[1],
 			       ntaylor,
@@ -195,6 +209,5 @@ expm <- function(x, method = c("Higham08.b", "Higham08",
 			       accuracy = double(1))[c("X", "accuracy")]
 	       structure(res$X, accuracy = res$accuracy)
 	   })## end{switch}
-
 }
 
